import os, sys, math, random, time, pygame
from pygame.locals import *
from pygame.compat import geterror

import const, util, image, item, bonus, paddle, interface, ball, explosion, weapon

global game

class Game():
    def __init__(self, screen, font):
        self.screen = screen
        self.font = font
        self.background = None
        self.initial_lives = 10
        self.lives = self.initial_lives
        self.level = 0
        self.initial_level = 0        #one below the actual initial level, it's incremented on restart
        self.score = 0
        self.combo = 0          #consecutive kills, raises the score
        self.enemies = 0
        self.bonus_delay = 0    #prevent bonuses falling in bunches if kills are too fast
        self.prev_bonus = -1  #to avoid repeating 2 same bonuses in a row
        self.max_bonus = 16   #more than size of bonuses{}, to decrease the number of bonuses by letting some kills go unrewarded
        self.bonuses = {0:"fire",
                        1:"balls",
                        2:"wide",
                        3:"net",
                        4:"freeze",
                        5:"sticky",
                        6:"life",
                        7:"strong",
                        8:"missile",
                        9:"bomb"
                        }
        self.bonus_image = {}   #preload images 'cause we need them all the time
        for b in self.bonuses:
            self.bonus_image[b] = util.load_image("images/bonus/" + self.bonuses[b] + ".png", -2)
        
        self.bonusdata = {"fire": {"cnt":0, "max":800, "func":self.use_fire},
                          "net": {"cnt":0, "max":2000, "func":self.use_net},
                          "freeze": {"cnt":0, "max":2000, "func":None},
                          "wide": {"cnt":0, "max":2000, "func":self.use_wide},
                          "strong": {"cnt":0, "max":2000, "func":self.use_strong},
                          "sticky": {"cnt":0, "max":2000, "func":None},
                          "missile": {"cnt":0, "max":2000, "func":self.use_missile},
                          "bomb": {"cnt":0, "max":300, "func":self.use_bomb},                          
                           }

        self.allsprites = pygame.sprite.LayeredDirty()
        self.balls = []
        self.bricks = []
        self.paddle = paddle.Paddle(self)
        self.net = bonus.Net()
        self.hud = interface.HUD(self)
        #create all available messages, easier than creating dynamically
        self.messages = {"main":interface.MainMenu(self),
                         "levelup":interface.LevelUp(self),
                         "gameover":interface.GameOver(self, False),
                         "youwin":interface.GameOver(self, True)
                         }
        self.cur_message = ''    #if non-empty, message is displayed instead of playfield
        self.bomb = None         #can be only one bomb
        self.bricks_todo = []   #dynamically created bricks, from mutable spiders
        self.paused = False
        
    def update_score(self, points):
        self.score += points
        self.hud.update_score()

    def update_combo(self):
        self.combo += 1
        self.hud.update_combo()
        self.update_score(2*self.combo)

    def clear_combo(self):
        self.combo = 0
        self.hud.update_combo()

    def lose_life(self):
        self.lives -= 1
        self.hud.update_lives()
        self.reinit()
        if self.lives <= 0 and not self.cheat:
            self.messages["gameover"].display()

    def restart(self):
        self.score = 0
        self.lives = self.initial_lives
        self.level = self.initial_level
        self.level_up()

    def reinit(self):
        self.bonus_delay = 0
        for b in self.bonusdata.keys():
            self.bonusdata[b]["cnt"] = 0
            if self.bonusdata[b]["func"]:
                self.bonusdata[b]["func"](False)
        self.paddle.init()
        for sprite in list(self.allsprites):
            if isinstance(sprite, ball.Ball) or isinstance(sprite, bonus.Bonus):
                self.allsprites.remove(sprite)
        self.balls = []    
        b = ball.Ball(True)
        self.balls.append(b)
        self.allsprites.add(b, layer=2)
        self.clear_combo()
        self.hud.reinit()
        self.hud.update_all()
        self.pause_shooting_spiders()
        self.bricks_todo = []    #for delayed creation of bricks from mutable spiders
        
    def level_done(self):
        if self.enemies > 0:
            return
        self.messages["levelup"].display()
            
    def level_up(self):
        self.level += 1
        self.bricks = []
        if self.load_level("levels/level" + str(self.level) + ".lvl"):
            return
        self.screen.blit(game.background, (0, 0))
        pygame.display.flip()
        self.allsprites.empty()
        self.allsprites.add(self.paddle, layer=2)
        self.allsprites.add(self.hud, layer=2)
        for brick in self.bricks:
            self.allsprites.add(brick, layer=brick.layer)
        self.allsprites.clear(game.screen, game.background)
        pygame.display.flip()
        self.reinit()

    def load_level(self, filename):
        try:
            f = open(filename, "r")
            background = f.readline().rstrip()   #remove newlines and spaces
            self.background = util.load_image("images/background/" + background, None)
            lines = f.readlines()
            f.close()
        except IOError:
            #file not found - we got past the last level, which means winning
            self.messages["youwin"].display()
            return 1
        row = 0
        n = 0    #n is for randomizing the spider graphics from the sprite sheet
        self.enemies = 0
        for line in lines:
            col = 0
            for c in line:
                x = col*50
                y = const.SCREEN_Y0 + row*50
                if c == 'z' or c == 'y' or c == 'x' or c == 'v':
                    self.bricks.append(item.Spider(x, y, n, c))
                    self.enemies += 1
                    n += 1
                elif c == 'w':
                    self.bricks.append(item.ShootingSpider(x, y, n, c))
                    self.enemies += 1
                    n += 1
                elif c == 'u':
                    self.bricks.append(item.MutableSpider(x, y, n, c))
                    self.enemies += 1
                    n += 1
                elif c == 'a':    
                    self.bricks.append(item.BreakableBrick(x, y, 3)) 
                elif c == 'b':
                    self.bricks.append(item.UnbreakableBrick(x, y))
                elif c == 'c':
                    self.bricks.append(item.MutableBrick(x, y))
                elif c == 'd':
                    self.bricks.append(item.DynamiteBrick(x, y))
                elif c == 'e':
                    self.bricks.append(item.FallingBrick(x, y, 3))
                elif c == 'f':
                    self.bricks.append(item.BubbleBrick(x, y, 2))
                col += 1
            row += 1
        return 0    

    def kill_all_spiders(self):    #the cheating function
        for brick in list(self.bricks):
            if brick.enemy:
                self.bricks.remove(brick)
                self.allsprites.remove(brick)
                self.enemies -= 1

    def add_bonus(self, x, y):     #generate the bonus upon killing 
        if self.bonus_delay > 0:
            return
        r = random.randrange(self.max_bonus)
        if r in self.bonuses:
            if r == self.prev_bonus:    #avoid repeating the same bonus twice in a row
                r = (r+1) % len(self.bonuses)
            self.prev_bonus = r    
            self.allsprites.add(bonus.Bonus(self.bonuses[r], self.bonus_image[r], (x, y) ))
            self.allsprites.clear(self.screen, self.background)
            self.bonus_delay = 400

    def use_bonus(self, obj):
        self.allsprites.remove(obj)
        name = obj.name
        #call a bonus-specific function
        if name == "balls":
            b = ball.Ball(False, pos=list(self.balls[0].pos), vel=[-2,2])
            if self.bonusdata["fire"]["cnt"] > 0:
                b.fire(True)
            self.balls.append(b)
            self.allsprites.add(b)
            self.allsprites.clear(self.screen, self.background)
        elif name == "life":
            self.lives += 1
            self.hud.update_lives()
        else:
            self.bonusdata[name]["cnt"] = self.bonusdata[name]["max"]
            if self.bonusdata[name]["func"]:
                self.bonusdata[name]["func"](True)

    def use_net(self, use):
        if use:
            self.allsprites.add(self.net, layer=2)
        else:
            self.allsprites.remove(self.net)             

    def use_fire(self, use):
        if use:
            for ball in self.balls:
                ball.fire(True)
        else:
            for ball in self.balls:
                ball.fire(False)

    def use_wide(self, use):
        if use:
            self.paddle.expand(1)
        else:
            self.paddle.expand(-1)

    def use_strong(self, use):
        if use:
            self.paddle.strengthen(True)
        else:
            self.paddle.strengthen(False)


    def use_bomb(self, use):
        if use:
            self.bomb = weapon.Bomb(self.bonusdata["bomb"]["max"])
            self.allsprites.add(self.bomb)

    def use_missile(self, use):
        pass
        #it lives its own life, is used by itself after it's released in update_missile

    def update_missile(self):
        if self.bonusdata["missile"]["cnt"] % 40 == 0:
            m = weapon.Missile(self.paddle.pos, self.paddle.rect[1])
            self.allsprites.add(m)

    def pause_shooting_spiders(self):
        for sprite in list(self.allsprites):
            if isinstance(sprite, weapon.SpiderMissile):
                self.allsprites.remove(sprite)
            elif isinstance(sprite, item.ShootingSpider):
                sprite.pause(True)

    def resume_shooting_spiders(self):
        for sprite in list(self.allsprites):
            if isinstance(sprite, item.ShootingSpider):
                sprite.pause(False)

    def update_bonuses(self):    #update the dials to show remaining time for bonuses
        if self.bonus_delay > 0:
            self.bonus_delay -= 1

        for bonus in self.bonusdata.keys():
            if self.bonusdata[bonus]["cnt"] > 0:
                self.bonusdata[bonus]["cnt"] -= 1
                if self.bonusdata[bonus]["cnt"] <= 0:  #check for <0 just in case
                    self.bonusdata[bonus]["cnt"] = 0
                    if self.bonusdata[bonus]["func"]:
                        self.bonusdata[bonus]["func"](False)
                    self.hud.update_bonus(bonus, 0)
                else:
                    self.hud.update_bonus(bonus, 100*self.bonusdata[bonus]["cnt"]//self.bonusdata[bonus]["max"])

        if self.bonusdata["missile"]["cnt"] > 0:
            self.update_missile()

    def dynamic_bricks(self):
        #bricks created from mutable spiders
        #wait till the ball is far enough from the position, otherwise ball ends up inside the new brick or hits it immediately
        for pos in list(self.bricks_todo):
            x = pos[0]
            y = pos[1]
            use = True
            for ball in self.balls:
                d = math.sqrt(math.pow(ball.pos[0] - x,2) + math.pow(ball.pos[1]-y,2))  #distance between the ball and the new brick location
                if d < 80:
                    use = False   #not now, one of the balls is still too close
                    break
            if use:
                r = random.randrange(0, 20)    #different probabilities for different types of bricks
                if r <= 10:
                    b = item.BreakableBrick(x, y, 3)
                elif r <= 14:
                    b = item.FallingBrick(x, y, 3)
                elif r <= 18:
                    b = item.BubbleBrick(x, y, 2)
                else:  #r==19
                    b = item.DynamiteBrick(x, y)
                self.bricks.append(b)
                self.allsprites.add(b)
                self.bricks_todo.remove(pos)

    def click_callback(self):
        for b in self.balls:
            if b.on_paddle:
                b.release()
                break
        if self.bomb:
            self.bomb.release()
        self.resume_shooting_spiders()

    def event_loop(self):
        for event in pygame.event.get():
            if event.type == QUIT:
                return False
            elif event.type == KEYDOWN:
                if (event.key == pygame.K_LEFT or event.key == pygame.K_RIGHT) and not self.paused:
                    self.paddle.key_down(event.key)                
                elif event.key == pygame.K_SPACE and not self.paused:
                    self.click_callback()
                elif event.key ==  pygame.K_DELETE and not self.balls[0].on_paddle and self.cheat and not self.paused:
                    self.kill_all_spiders()
                elif event.key == pygame.K_p:
                    self.paused = True
                elif event.key == pygame.K_o:
                    self.paused = False

            elif event.type == KEYUP and not self.paused:
                self.paddle.key_up(event.key)
            elif event.type == MOUSEMOTION and not self.paused:
                self.paddle.mouse_move(event.pos)
            elif event.type == MOUSEBUTTONDOWN and not self.paused:
                self.click_callback()

        if self.paused:
            return True
        
        if len(self.bricks_todo) > 0:
            self.dynamic_bricks()

        self.allsprites.update()
        self.update_bonuses()
        rects = self.allsprites.draw(self.screen)
        pygame.display.update(rects)
        
        self.level_done()   #check if level is finished
        return True

    def play(self):   #main game loop
        self.level_up()
        clock = pygame.time.Clock()
        going = True
        self.messages["main"].display()
        while going:
            clock.tick(60)
            if self.cur_message != '':
                going = self.messages[self.cur_message].event_loop()
            else:
                going = self.event_loop()
            pygame.display.flip()
            
        

